##########################################################################
# This is an example and debug protocol file for StreamDevice.
#
# (C) 2010 Dirk Zimoch (dirk.zimoch@psi.ch)
#
# This file is part of StreamDevice.
#
# StreamDevice is free software: You can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StreamDevice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with StreamDevice. If not, see https://www.gnu.org/licenses/.
#########################################################################/

terminator = LF;
readtimeout = 10;
pollperiod = 10;
replytimeout = 1000;
command {
    out "%s";
}
info {
    out "%s";
    in "%39c";
}
request {
    out "%s";
    in "%(\$1)39c";
}
spy {
    extraInput=ignore;
    in "%39c";
}
spybin {
    readtimeout = 1;
    terminator = "";
    extraInput=ignore;
    in "%r";
}
checksum {
    out "%s %0.1<\$1>";
}
